#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Code common to the ry* series of command-line tools.

This module is wholly supplanted by rytools, except for the deprecated
mkcb_tickspec. It should be removed in Rayon 2.0
"""
from __future__ import division

import re

import miscutils
import datetime
import netsa
import socket
import struct
import optparse

from rayon import RayonException


# Option processing code

def mk_tickspec_parser(spec):
    """
    Convert a string specification of where to put tickmarkers into
    series of tick factories that return tick locations appropriate to
    the specification given data.

    @type  spec: string
    @param spec: comma-separated list of tick specifications. Valid
        values are C{min}, C{median}, and C{max}, which put ticks at
        the minimum, median and maximum values in the data,
        respectively. Another valid value is C{pN}, where I{N} is a
        value from 1 to 99 representing a percentile (e.g., C{median}
        is equivalent to C{p50}). Finally, the value C{nN} will put a
        tickmark at the numeric position I{N} (e.g., C{n50} puts a
        tickmark at the value 50).
    """
    def cbmin(col):
        return [col.min()]
    cbmin.label = "min"
    def cbmax(col):
        return [col.max()]
    cbmax.label = "max"
    def mkcbhighpass(thresh, inclusive):
        if inclusive:
            def cbhighpass(col):
                return (c for c in col if col >= thresh)
        else:
            def cbhighpass(col):
                return (c for c in col if col > thresh)
        return cbhighpass
    def mkcblowpass(thresh, inclusive):
        if inclusive:
            def cblowpass(col):
                return (c for c in col if col <= thresh)
        else:
            def cblowpass(col):
                return (c for c in col if col < thresh)
        return cblowpass
    def mkcbnice(num_ticks):
        def cbnice(col):
            if isinstance(col[0], datetime.datetime):
                lo, hi, nice_ticks = netsa.data.nice.nice_time_ticks(
                    col.min(), col.max(), num_ticks)
            else:
                lo, hi, nice_ticks = netsa.data.nice.nice_ticks(
                    col.min(), col.max(), num_ticks)
            return nice_ticks
        return cbnice
    def mkcbpercentile(num):
        def cbpercentile(col):
            return [col.percentile(num)]
        return cbpercentile

    if spec == "none" or spec == "":
        return []
    
    toks = spec.split(",")
    parsers = []
    for t in toks:
        if t.startswith("p"):
            pct = int(t[1:])
            f = mkcbpercentile(pct)
            f.label = "p%d" % pct
            parsers.append(f)
        elif t.startswith("n"):
            numticks = int(t[1:])
            f = mkcbnice(numticks)
            f.label = "n%d" % numticks
            parsers.append(f)
        elif t.startswith(">="):
            thresh = int(t[2:])
            f = mkcbhighpass(thresh, True)
            f.label = ">=%d" % thresh
            parsers.append(f)
        elif t.startswith(">"):
            thresh = int(t[:])
            f = mkcbhighpass(thresh, False)
            f.label = ">%d" % thresh
            parsers.append(f)
        elif t.startswith("<="):
            thresh = int(t[2:])
            f = mkcblowpass(thresh, True)
            f.label = "<=%d" % thresh
            parsers.append(f)
        elif t.startswith("<"):
            thresh = int(t[:])
            f = mkcblowpass(thresh, False)
            f.label = "<%d" % thresh
            parsers.append(f)
        elif t.lower() == "min":
            parsers.append(cbmin)
        elif t.lower() == "median":
            f = mkcbpercentile(50)
            f.label = "median"
            parsers.append(f)
        elif t.lower() == "max":
            parsers.append(cbmax)
        else:
            raise ValueError("Unexpected tickmark value: '%s'" % t)
    return parsers


# Optparse options processor callback factories

def mkcb_input(dst_val):
    """
    Make an optparse callback function that converts a
    L{rayon.data.Dataset} column specifier to an integer or string, as
    appropriate.

    @type  dst_val: string
    @param dst_val: The name of the option being handled with the
        callback
    """
    def parse_input(option, opt_str, value, parser):
        try:
            in_num = int(value)
            setattr(parser.values, dst_val, in_num)
        except ValueError:
            setattr(parser.values, dst_val, value)
    return parse_input


def mkcb_ticks(dst_val):
    """
    Make an optparse callback function that converts a tick
    specification to a set of tick-producing functions.

    @type  dst_val: string
    @param dst_val: The name of the option being handled with the
        callback
    """
    def parse_input(option, opt_str, value, parser):
        try:
            setattr(parser.values, dst_val, mk_tickspec_parser(value))
        except ValueError, e:
            raise optparse.OptionValueError(str(e))
    return parse_input


def mkcb_args(dst_val):
    """
    Make an optparse callback function that converts a set of
    comma-separated key=value pairs into a dictionary. The types of
    the dictionary values are inferred from the argument.

    @type  dst_val: string
    @param dst_val: The name of the option being handled with the
        callback
    """
    def parse_input(option, opt_str, value, parser):
        setattr(parser.values, dst_val,
                dict((k, miscutils.infer_type_from_string(v)(v)) for k, v in
                     [sa.split("=") for sa in value.split(",")]))
    return parse_input


def mkcb_color(dst_val):
    """
    Make an optparse callback function that converts a color
    specification into a 3- (RGB) or 4-tuple (RGBA) of ints from
    0-255.

    This function passes I{dst_val} directly to
    L{miscutils.parse_color_string}. However, unlike
    C{parse_color_string}, a hex RGB or RGBA specification may be
    passed without an initial hash mark. (This simplifies command-line
    invocation.)

    @type  dst_val: string
    @param dst_val: The name of the option being handled with the
        callback
    """
    def parse_input(option, opt_str, value, parser):
        try:
            setattr(parser.values,
                    dst_val,
                    miscutils.parse_color_string(value))
        except ValueError, e:
            # Try with a hash mark prepended
            try:
                setattr(parser.values,
                        dst_val,
                        miscutils.parse_color_string("#" + value))
            except:
                # Raise original error
                raise e
    
    return parse_input


def mkcb_intfromip(dst_val):
    """
    Make an optparse callback function that converts an integer or
    dotted-quad IPv4 address to an integer.
    """
    def parse_input(option, opt_str, value, parser):
        try:
            setattr(parser.values, dst_val, int(value))
        except ValueError:
            setattr(parser.values,
                    dst_val,
                    struct.unpack("!L", socket.inet_aton(value))[0])
    return parse_input
                    

    


