#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Code for serializing and deserializing L{Dataset} objects.
"""
import miscutils, csv
from copy import copy
import itertools

from rayon import RayonException

class DataParseException(RayonException):
    # Internal base class of exceptions in this module -- should
    # always be caught and handled before reaching a user.
    pass

class TypeConversionException(DataParseException):
    # Problem converting a value to the requested type in the typemap
    pass


class PushbackReader(object):
    """
    A read-only stream that supports pushing data back onto the stream
    after it's been read.

    *stream* is an object with a readline method.

    The last line read from this object may be pushed back into the
    stream with the push_back() method. To read the next line without
    advancing the stream, use peek().
    """
    def __init__(self, stream):
        self.stream = stream
        self.line = None
        self.buffer_is_live = False
    def _readline(self):
        if self.buffer_is_live:
            self.buffer_is_live = False
            line = self.line
            return line
        else:
            self.line = self.stream.readline()
            return self.line
    def readline(self, skipping=[]):
        """
        Read the next line of text that "should" be returned.

        *skipping* is a sequence of predicate functions that take a
        line of text and return ``True`` if the text should be
        skipped, otherwise ``False``.  If any of the predicates
        returns ``True``, the line will not be returned.
        """
        # Someday we'll no longer support python 2.4 and this can go away:
        def any(iterable):
            for i in iterable:
                if i:
                    return True
            return False
        while True:
            line = self._readline()
            if line == '': # EOF, always return
                return line
            if any(skip(line) for skip in skipping):
                continue
            return line
    def pushback(self):
        """
        Put the last line of text returned by the :meth:`readline`
        method back on the stream.
        """
        self.buffer_is_live = True
    def peek(self, skipping=[]):
        """
        Read the next line of text without advancing the stream.

        *skipping* is a sequence of predicate functions that take a
        line of text and return ``True`` if the text should be
        skipped, otherwise ``False``. If any of the predicates returns
        ``True``, the line will not be returned.
        """
        line = self.readline(skipping)
        self.pushback()
        return line
    def line_iter(self, skipping=[]):
        """
        Iterate through the lines in the stream until they are gone.

        *skipping* is a sequence of predicate functions that take a
        line of text and return ``True`` if the text should be
        skipped, otherwise ``False``. If any of the predicates returns
        ``True``, the line will not be returned.
        """
        while True:
            line = self.readline(skipping=skipping)
            if line == '':
                break
            yield line

        

# Predicates used in skipping parameter to readline() and peek()

    
def blank_p(line):
    """
    Returns ``True`` if *line* is non-EOF whitespace, otherwise
    ``False``.
    """
    return line.strip() == ''


def comment_p(line):
    """
    Returns ``True`` if *line* is a comment (i.e., it starts with the
    comment character, "#"). Otherwise ``False``.
    """
    return line.startswith('#')

def header_p(line):
    """
    Returns ``True`` if *line* is a header (i.e., it starts with 2
    comment characters and contains a colon). Otherwise
    ``False``. Note that if *line* passes :func:`header_p`, it will
    also pass :func:`comment_p`.
    """
    return line.startswith('##') and ":" in line

def blank_header_p(line):
    """
    Returns ``True`` if *line* is an empty header. (i.e., only
    whitespace). Otherwise ``False``.
    """
    return header_p(line) and line[2:].strip() == ''





def type_from_string(spec):
    # Generate appropriate typemaps for data with typemaps in
    # headers. Try to do the right thing for converters in as-yet
    # unimported modules.
    if "." in spec:
        pivot = spec.rfind(".")
        mod_name = spec[:pivot]
        fn_name = spec[pivot+1:]
        # fn_name is irrelevant in this call -- it can be any string
        # and still work (but it won't work without the arg -- see
        # __import__ docs)
        mod = __import__(mod_name, globals(), locals(), fn_name)
        return getattr(mod, fn_name)
    else:
        # Assume it's a builtin
        return __builtins__[spec]



class StreamParser(object):
    def __init__(self, colnames, header, typemap, rows, delimiter):
        self.__colnames = colnames
        self.__header = header
        self.__typemap = typemap
        self.__rows = rows
        self.__delimiter = delimiter
        self.__numcols = None
    def colnames(self):
        if self.__colnames is None:
            return None
        # Only return column names we have typeable data for
        return self.__colnames[:self.numcols()]
    def header(self): return self.__header
    def numcols(self):
        if self.__numcols is None:
            # Find the first line of the dataset and get the number of
            # columns from that.
            line = self.__rows.peek(skipping=(blank_p, comment_p))
            numcols = len(line.split(self.__delimiter))
            # Take the smallest of the number of typemaps and the
            # number of columns, and ignore the rest. If you have more
            # columns than typemaps, that means data will be ignored.
            self.__numcols = min(numcols, len(self.__typemap))
        return self.__numcols
    def row_iter(self):
        # If we got more typemaps than data columns, ignore the extras
        typemap = self.__typemap[:self.numcols()]

        for line in self.__rows.line_iter(skipping=(blank_p, comment_p)):
            items = line.split(self.__delimiter)[:self.numcols()]
            tmp = []
            def convert(fn, inval):
                try:
                    return fn(inval)
                except ValueError, e:
                    raise TypeConversionException(fn.__name__, inval, e)
            yield [convert(typemap[idx], i.strip())
                   for idx, i in enumerate(items)]

class EmptyStreamParser(object):
    # parser for empty input files. In this case, all the columns
    # listed in colnames exist, have type 'str' (this should probably
    # change), and no data
    def __init__(self, colnames, header):
        self.__colnames = colnames
        self.__header = header
    def colnames(self):
        if self.__colnames is None:
            return None
        else:
            return self.__colnames
    def header(self): return self.__header
    def numcols(self):
        # Rules are a little different for empty datasets. There are
        # no values, so type maps aren't a problem. Return the number
        # of columns reported by the column names, if any, otherwise 0
        if self.colnames() is None:
            return 0
        else:
            return len(self.colnames())
    def row_iter(self):
        return iter(tuple())






##################################################
# from_stream and its helper functions
##################################################



def _decipher_column_names(user_colnames,
                          hdr_colnames,
                          oldskool_colnames,
                          delimiter):
    # Column name spec precedence:
    # user_colnames > column-names header > old-style column names > none
    if user_colnames is not None:
        return tuple(n.strip() for n in user_colnames)
    if hdr_colnames is not None:
        return tuple(n.strip() for n in hdr_colnames.split(delimiter))
    if (oldskool_colnames is not None
        and delimiter is not None
        and delimiter in oldskool_colnames):
        return tuple(n.strip() for n in oldskool_colnames.split(delimiter))
    return None

def _decipher_typemap(user_typemap, hdr_typemap, first_row, delimiter):
    # Typemap precedence:
    # user_typemap > typemap header > inferred typemap
    if user_typemap is not None:
        return user_typemap
    if hdr_typemap is not None:
        if delimiter is None:
            # Assume a single column
            return (type_from_string(hdr_typemap.strip()),)
        else:
            return tuple(type_from_string(s.strip())
                         for s in hdr_typemap.split(delimiter))
    if first_row is not None:
        if delimiter is None:
            return (miscutils.infer_type_from_string(first_row.strip()),)
        else:
            return tuple(miscutils.infer_type_from_string(s.strip())
                         for s in first_row.strip().split(delimiter))
    return tuple()

def _decipher_delimiter(user_delimiter, hdr_delimiter):
    # Delimiter precedence:
    # user_delimiter > delimiter header > default
    if user_delimiter is not None:
        return user_delimiter
    if hdr_delimiter is not None:
        return hdr_delimiter
    return '|'


def _strip_comment_chars(line):
    line = line.lstrip()
    while line.startswith('#'):
        line = line[1:]
    return line


def _read_header(stream):
    def hdr_line_gen():
        # Skip blank lines (and blank headers), but not comments
        for line in stream.line_iter(skipping=(blank_p, blank_header_p)):
            if not header_p(line):
                stream.pushback()
                break
            line = _strip_comment_chars(line)
            k, v = line.split(":", 1)
            yield (k.lower().strip(), v.strip())
    return dict(h for h in hdr_line_gen())


def _read_comments(stream):
    # Read and return any comment lines that are at the top of the
    # stream. Treats header comments as comments---the header
    # comments should have been read by the time read_comments is
    # called.
    def comment_line_gen():
        for line in stream.line_iter(skipping=(blank_p,)):
            if not comment_p(line):
                stream.pushback()
                break
            yield _strip_comment_chars(line).strip()
    return tuple(l for l in comment_line_gen())
        

def from_stream(stream, 
                orig_colnames, 
                orig_typemap, 
                orig_delimiter,
                first_line_is_colnames):
    """Used by L{Dataset.from_stream} to marshall data into a L{Dataset}."""
    stream = PushbackReader(stream)

    header = _read_header(stream)
    # Fetch any comments between the header and the data. The last
    # line in this block (if it exists) may be an old-style column
    # specifier.
    comment_block = _read_comments(stream)
    if len(comment_block) > 0:
        oldskool_colnames = comment_block[-1]
    else:
        oldskool_colnames = None

    delimiter = _decipher_delimiter(orig_delimiter,
                                    header.get('delimiter', None))
    colnames = _decipher_column_names(orig_colnames,
                                     header.get('column-names', None),
                                     oldskool_colnames,
                                     delimiter)

    # If all else fails (and we're told to), check for silk-style
    # column names in first line of input.
    if colnames is None and first_line_is_colnames:
        colnames_row = stream.readline()
        colnames = tuple(s.strip() for s in colnames_row.split(delimiter))
            
        

    first_row = stream.peek(skipping=(comment_p, blank_p))

    typemap = _decipher_typemap(
        orig_typemap, header.get('typemap', None), first_row, delimiter)

    # Remove the meta-information headers. We'll recreate them if we
    # reserialize.
    for h in ('column-names', 'delimiter', 'typemap'):
        if header.has_key(h):
            del header[h]


    # Is dataset empty?
    if first_row == '':
        return EmptyStreamParser(colnames, header)
    else:
        return StreamParser(colnames, header, typemap, stream, delimiter)

    
    






def to_stream(stream, colnames, header, row_iter, numrows, delimiter):
    """Used by L{Dataset.to_stream} to serialize a Dataset to a stream."""
    for k, v in header.items():
        stream.write("## %s: %s\n" % (k, v))
    if len(colnames) > 0:
        stream.write("## column-names: %s\n" % delimiter.join(colnames))
    for idx, row in enumerate(row_iter):
        stream.write(delimiter.join(str(e) for e in row))
        # Omit the final newline
        if idx != numrows - 1:
            stream.write("\n")



def from_csv_stream(stream, colnames=None, typemap=[], dialect=None):
    """Used by L{Dataset.from_csv_file} to marshall data into a L{Dataset}."""

    pos = stream.tell()
    sample = stream.read(1024)
    stream.seek(pos)

    if not sample:
        return stream, colnames

    wilbur = csv.Sniffer()
    wilbur.preferred.append('|')

    if dialect is None:
        dialect = wilbur.sniff(sample)

    has_colnames = wilbur.has_header(sample)

    stream = PushbackReader(stream)
    reader = csv.reader(stream, dialect)
            
    if has_colnames:
        try:
            cnames = []
            for i,x in enumerate(reader.next()):
                x = x.strip()
                cnames.append(x or i)
        except StopIteration:
            pass
    else:
        cnames = None

    if colnames:
        if len(colnames) != len(cnames):
            raise ValueError, "given colnames mismatch found %s" % str(cnames)
    else:
        colnames = cnames
        
    for row in reader:
        if row:
            row = [row[x] for x in xrange(len(colnames))]
            break
    stream.pushback()

    typemap_orig = typemap
    typemap = []
    for i,c in enumerate(colnames):
        c = c.strip()
        try:
            typemap.append(typemap_orig[i])
        except (IndexError, KeyError, TypeError):
            try:
                typemap.append(typemap_orig[c])
            except (IndexError, KeyError, TypeError):
                typemap.append(miscutils.infer_type_from_string(row[i].strip()))

    reader = ([typemap[i](r[i].strip())
                for i in xrange(len(colnames))] for r in reader)

    return reader, colnames


__all__ = ['from_stream', 'to_stream', 'from_csv_stream', 'DataParseException']
