#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

from __future__ import division

from collections import Iterable
from abc import ABCMeta, abstractmethod
from types import NoneType

from rayon.backcompat import OrderedDict
from rayon.data import Dataset

""".

A set of classes for manipulating data in a single
:class:`~rayon.data.Dataset` object, potentially into several labeled
data sets.

An example::

   d1 = Dataset.from_rows([["a", 1, "alpha"],
                           ["a", 2, "beta"],
                           ["b", 1, "gamma"],
                           ["b", 2, "delta"]],
                          colnames="x y z".split())

   def group_1(row):
       return row[0]

   def group_2 (row):
       if row[1] % 2 == 0: 
           return 'even'
       return 'odd'

   # Subdivide input into a Dataset for each value of column 0, then
   # further subdivide those into datasets for the even and odd numbers.
   chain = Classify(group_1) + Classify(group_2)

   # Run the chain on d1
   output = chain(d1)

   # Output is:
   #
   # OrderedDict([
   #     ('a', OrderedDict([
   #         ('odd', Dataset.from_rows([["a", 1, "alpha"]])),
   #         ('even',Dataset.from_rows([["a", 2, "beta"]]))])), 
   #     ('b', OrderedDict([
   #         ('odd', Dataset.from_rows([["b", 1, "gamma"]])), 
   #         ('even',Dataset.from_rows([["b", 2, "delta"]]))]))])
   
The goal of this package is to simplify a frequent use case in data
manipulation, especially for visualization: preparing a monolithic
dataset for presentation by extracting subsets, removing extraneous
rows/columns, or directly transforming the data itself. This is
usually done in a series of maps over the data, and the result is
usually a collection of smaller datasets that will be presented to the
user in some way.

Each :class:`Transformer` object is a callable that takes a
:class:`Dataset` object and returns either another :class:`Dataset`
object or an :class:`OrderedDict` object. The ordered dictionary
values are either :class:`Dataset` objects or other ordered
dictionaries with the same restriction. IOW, the return value is
either a dataset or a tree of datasets.

A chain of transformers is formed by adding one or more
:class:`Transformer` objects together. The result is another
:class:`Transformer` object which, when run on a dataset, passes the
dataset to the first link in the chain. Subsequent links in the chain
are run on each :class:`Dataset` object in the return value of the
previous chain, replacing each one in the tree as it goes along. Thus,
each new transform operation successively elaborates the tree,
replacing each leaf node with either another leaf node or a new
subtree.

"""


def _walk(node, action):
    # Walk a tree of Datasets, replacing each Dataset with the result
    # of *action*. *action* takes a dataset and a dictionary of state
    # information, and returns a tree node (a Dataset or an
    # OrderedDict).
    if isinstance(node, Dataset):
        return action(node, {'path': []})

    branches  = [(node, [])] # [branch_contents, path_stack]
    while len(branches) > 0:
        branch, pathinfo = branches.pop()
        for k, v in branch.items():
            if isinstance(v, Dataset):
                branch[k] = action(v, {'path': pathinfo + [k]})
            else:
                branches.append((v, pathinfo + [k]))
    return node
            
        
# Note: Yes, it's a monad. But it's kind of an a-hole move to call
# something Return in a language where return is a keyword, so I'm
# calling it Lift.

class Transformer:
    # Abstract base class of all Transformers
    __metaclass__ = ABCMeta
    def __add__(self, other):
        return Bind(self, other)
    def __call__(self, dset):
        return self.generate_children(dset, {'path': []})
    @abstractmethod
    def generate_children(self, dset, state):
        pass


class Lift(Transformer):
    """.  

    Perform a generic user-supplied transformation on datasets.
    *lift_fn* takes a :class:`Dataset` object and a state dictionary,
    and returns either another :class:`Dataset` object or an
    :class:`OrderedDict` object whose values are either
    :class:`Dataset` objects or other :class:`OrderedDict` objects
    with the same restriction. (IOW, a tree of datasets.)

    The state dictionary contains additional information about the
    current state of the computation. At the moment, the only field
    provided in the dictionary is ``path``, which contains an array of
    tree labels. The last element in the array is the label of the
    node *lift_fn* will return; the element before that is the label
    of this node's parent, the one before that its grandparent, and so
    on. The root node has no label, so the depth of the tree is
    len(state['path']) + 1.

    If supplied, and if the result of the transformation is a subtree
    (i.e., an :class:`OrderedDict` object), *ordering* is a
    representation of the order of labels (i.e., keys) in the top
    level of the subtree. If *ordering* is an iterable of potential
    subtree labels, every label in the subtree will appear in the
    order specified in *ordering*. *ordering* must contain every label
    in the subtree. *ordering* may contain additional labels; they
    will be ignored. *ordering* may optionally be a function that
    takes an :class:`OrderedDict` object representing the subtree and
    returns an iterable, as previously described. This function must
    have no side effects; in particular, this function must not mutate
    its input value in any way.

    If supplied, *name* is the name of the transformation operation,
    and will be used in error reporting.

    """
    def __init__(self, lift_fn, ordering=None, name=None):
        self.lift_fn = lift_fn
        if ordering is None:
            self.ordering_fn = None
        elif callable(ordering):
            self.ordering_fn = ordering
        else:
            # Hope it's a sequence
            def _ordering_fn(odict):
                return ordering
            self.ordering_fn = _ordering_fn
        self.name = name
    def generate_children(self, dset, state):
        subtree = self.lift_fn(dset, state)
        if self.ordering_fn is not None and isinstance(subtree, OrderedDict):
            # Reorder subtree labels
            subtree = OrderedDict([(k, subtree[k]) 
                                   for k in self.ordering_fn(subtree)
                                   if k in subtree])
        return subtree
        

class Bind(Transformer):
    # The composition operation for Transformers
    def __init__(self, xform1, xform2):
        # To avoid recursion, unbox any Bind objects that get passed in.
        ops1 = xform1.ops if isinstance(xform1, Bind) else [xform1]
        ops2 = xform2.ops if isinstance(xform2, Bind) else [xform2]
        self.ops = ops1 + ops2

    def generate_children(self, dset, state):
        def reducer(tree, nxt):
            return _walk(tree, nxt.generate_children) 

        return reduce(reducer, 
                      self.ops[1:], 
                      self.ops[0].generate_children(dset, state))



def _Classify(pass_state, classifier_fn, ordering=None):
    def generate_children(dset, state):
        children = OrderedDict()
        for row in dset:
            if pass_state:
                labels = classifier_fn(row, state)
            else:
                labels = classifier_fn(row)
            
            if labels is not None:
                # Allow users to specify a single return value if row
                # belongs in only one group. (But for user
                # convenience, treat strings as single values, not
                # iterables.)
                if (isinstance(labels, basestring) 
                    or not isinstance(labels, Iterable)):
                    labels = [labels]

                for label in labels:
                    children[label] = children.get(label, []) + [row]
        # Convert children to Datasets before returning
        def to_dataset(rows, colnames):
            if len(colnames) == 0:
                return Dataset.from_rows(rows)
            else:
                return Dataset.from_rows(rows, colnames)
        return OrderedDict([(k, to_dataset(v, dset.get_column_names()))
                          for k, v in children.items()])
    return Lift(generate_children, ordering=ordering)

def Classify(classifier_fn, ordering=None):
    """
    Classify the incoming data into groups. *classifer_fn* takes a row
    and returns either a single label, an iterable collection of
    labels, or ``None`` (indicating the row should be discarded; this
    is equivalent to returning an empty collection). The result of
    this transformation is an ordered dictionary whose keys are the
    labels returned by *classifier_fn* and whose values are
    :class:`Dataset` objects containing the rows for which
    *classifier_fn* returned that label.

    A "label" is defined as a hashable value. When returning a single
    hashable value that is also iterable, return it as a one-element
    list or tuple. (E.g., as ``[x]`` instead of ``x``.) The exception
    to this is strings; when strings are returned from
    *classifier_fn*, they are treated as unitary values, not
    collections. (Return collections of characters as a list or other
    datatype. For instance, use ``['a', 'b', 'c']`` instead of
    ``"abc"``).

    If supplied, *ordering* is the order of labels in the subtree, as
    described in the :class:`Lift` object.
    """
    return _Classify(False, classifier_fn, ordering=ordering)

def StClassify(classifier_fn, ordering=None):
    """
    Identical to the :func:`Classify` operation, but *classifier_fn*
    takes a row and a dictionary of state information, as with the
    :class:`Lift` operation.
    """
    return _Classify(True, classifier_fn, ordering=ordering)


def _Project(pass_state, projection_fn, colnames):
    def generate_children(dset, state):
        def mapper(row):
            if pass_state:
                return projection_fn(row, state)
            else:
                return projection_fn(row)
        return Dataset.from_rows(map(mapper, dset), colnames=colnames)
    return Lift(generate_children)

def Project(projection_fn, colnames):
    """
    Project incoming datasets (in the relational
    sense). *projection_fn* takes a row and returns a tuple that will
    form a new row based on it. The result of this transformation is a
    :class:`Dataset` object whose rows are the tuples returned from
    *projection_fn*.

    If the manipulations to be done by *project_fn* only remove or
    reorder columns, consider using the :class:`Reorder` transform,
    which does this much more efficiently.
    """
    return _Project(False, projection_fn, colnames)

def StProject(projection_fn, colnames):
    """
    Identical to the :func:`Project` operation, but *projection_fn*
    takes a row and a dictionary of state information, as with the
    :class:`Lift` operation.
    """
    return _Project(True, projection_fn, colnames)


def _Reorder(pass_state, new_colnames):
    # Normalize to a function that takes a dataset and state
    if callable(new_colnames):
        if pass_state:
            new_colnames_fn = new_colnames
        else:
            new_colnames_fn = lambda dset, state: new_colnames(dset)
    else:
        new_colnames_fn = lambda dset, state: new_colnames
        
    def generate_children(dset, state):
        return Dataset.from_columns(
            [dset.get_column(i) for i in new_colnames_fn(dset, state)],
            colnames=new_colnames_fn(dset, state))
    return Lift(generate_children)

def Reorder(new_colnames):
    """.

    Reorder (or remove) the columns in datasets. *new_colnames* is
    either a sequence of column names or indices, or a function that
    takes a :class:`Dataset` object and returns a sequence of column
    names or indices. Each name/index in *new_colnames* (or it's
    output, if it's a function) must correspond to one in the incoming
    :class:`Dataset` object. The result of this transformation is a
    new dataset based on the incoming dataset, with only the columns
    mentioned in *new_colnames*, in the order specified in
    *new_colnames*.

    For more involved row-based transformations, see the
    :class:`Project` operation.

    """
    return _Reorder(False, new_colnames)

def StReorder(new_colnames):
    """
    Identical to the :func:`Reorder` operation, but if *new_colnames*
    is a function, it takes a :class:`Dataset` object and a dictionary
    of state information, as with the :class:`Lift` operation.
    """
    return _Reorder(True, new_colnames)

def _Include(pass_state, include_fn):
    def generate_children(dset, state):
        if pass_state:
            pred = lambda row: include_fn(row, state)
        else:
            pred = include_fn
        return dset.filter_pass(pred)
    return Lift(generate_children)

def Include(include_fn):
    """
    Remove rows from datasets. *include_fn* takes a row and returns
    ``True`` or ``False``; if it returns ``True`` the row will be
    included in the result set. The result of the transformation is a
    new dataset containing only the rows for which *include_fn*
    returned ``True``.
    """
    return _Include(False, include_fn)

def StInclude(include_fn):
    """
    Identical to the :func:`Include` operation, but *include_fn*
    takes a row and a dictionary of state information, as with the
    :class:`Lift` operation.
    """
    return _Include(True, include_fn)



def _Exclude(pass_state, exclude_fn):
    if pass_state:
        pred = lambda row: exclude_fn(row, state)
    else:
        pred = exclude_fn
                 
    def generate_children(dset, state):
        return dset.filter_fail(pred)
    return Lift(generate_children)


def Exclude(exclude_fn):
    """
    Remove rows from datasets. *exclude_fn* takes a row and returns
    ``True`` or ``False``; if it returns ``True`` the row will be
    excluded from the result set. The result of the transformation is a
    new dataset containing only the rows for which *include_fn*
    returned ``False``.
    """
    return _Exclude(False, exclude_fn)

def StExclude(exclude_fn):
    """
    Identical to the :func:`Exclude` operation, but *exclude_fn*
    takes a row and a dictionary of state information, as with the
    :class:`Lift` operation.
    """
    return _Exclude(False, exclude_fn)
        

__all__ = """
Bind
Classify
Exclude
Include
Lift
Project
Reorder
StClassify
StExclude
StInclude
StProject
StReorder
Transformer
""".split()
