#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@

from __future__ import division

import math

"""
Fundamental geometric abstractions used throughout Rayon and tools for
operating on and converting between them. This is currently an
internal API and should not be publicly documented or used. However,
it may be provided in the future to those wishing to extend Rayon in
ways that use the 2D drawing API in the :mod:`rayon.backend` module.

Highlights of this library:

   * The :class:`Point` object and subclasses, to represent different
     types of geometric points. The primary purpose of these objects
     is as tags, so values in different coordinate systems aren't
     accidentally mixed. They may be treated as tuples to get their
     constituent values; they can be compared with other objects of
     the same type for equality. :class:`CartesianPoint` and
     :class:`ScreenPoint` objects can be added to and subtracted from
     other objects of the same type (by adding/subtracting their ``x``
     and ``y`` components).

   * Functions for converting points between coordinate
     systems. Functions exist to convert directly between
     :class:`PolarPoint` objects and :class:`CartesianPoint' objects,
     and between :class:`CartesianPoint` objects and
     :class:`ScreenPoint` objects. Conversion between
     :class:`ScreenPoint` objects and :class:`PolarPoint` objects
     requires an intermediate conversion to :class:`CartesianPoint`
     objects.

   * The :class:`Rectangle` object, which denotes a rectangular region
     of space. The :class:`Rectangle` object is intended for use with
     the 2D API of the :mod:`rayon.backend` module; as such, it deals
     exclusively in :class:`ScreenPoints`.

Units
-----

The classes in this module assume all measures have been converted
into the unit space used by the 2D drawing API (points, pixels,
etc.). The :mod:`rayon.backend` module contains classes and methods
for specifying and converting between other units of measure.

"""


#
# Points
#

class Point(object):
    """
    Base class for various types of points. The implemented subtypes
    of this class are:

       * :class:`CartesianPoint`
       * :class:`ScreenPoint`
       * :class:`PolarPoint`
    """
    pass

class CartesianPoint(Point):
    """
    A point on a classical Cartesian plane. (i.e., y increases going
    up). The ``x`` and ``y`` components of the point may be accessed
    as with any sequence::

       # pt is a CartesianPoint
       x = pt[0]
       y = pt[1]

       # or

       x, y = pt
    """
    def __init__(self, x, y):
        self._x = x
        self._y = y
    def __getitem__(self, key):
        if key == 0: return self._x
        elif key == 1: return self._y
        else: raise IndexError(key)
    def __eq__(self, other):
        if (isinstance(other, CartesianPoint) and
            self._x == other._x and
            self._y == other._y):
            return True
        return False
    def __add__(self, other):
        if not isinstance(other, CartesianPoint):
            raise TypeError(
                "unsupported operand type(s) for +: '%s' and '%s'" %
                (type(self).__name__, type(other).__name__))
        return CartesianPoint(self._x + other._x, self._y + other._y)
    def __sub__(self, other):
        if not isinstance(other, CartesianPoint):
            raise TypeError(
                "unsupported operand type(s) for -: '%s' and '%s'" %
                (type(self).__name__, type(other).__name__))
        return CartesianPoint(self._x - other._x, self._y - other._y)
    def __repr__(self):
        return "CartesianPoint(%s, %s)" % (self._x, self._y)

class ScreenPoint(Point):
    """
    Like a :class:`CartesianPoint`, but y increases going down. This
    is the point used when dealing with screen coordinates, where the
    origin is in the top left.

    The ``x`` and ``y`` components of the point may be accessed as
    with any sequence::

       # pt is a ScreenPoint
       x = pt[0]
       y = pt[1]

       # or

       x, y = pt    
    """
    def __init__(self, x, y):
        self._x = x
        self._y = y
    def __getitem__(self, key):
        if key == 0: return self._x
        elif key == 1: return self._y
        else: raise IndexError(key)
    def __eq__(self, other):
        if (isinstance(other, ScreenPoint) and
            self._x == other._x and
            self._y == other._y):
            return True
        return False
    def __add__(self, other):
        if not isinstance(other, ScreenPoint):
            raise TypeError(
                "unsupported operand type(s) for +: '%s' and '%s'" %
                (type(self).__name__, type(other).__name__))
        return ScreenPoint(self._x + other._x, self._y + other._y)
    def __sub__(self, other):
        if not isinstance(other, ScreenPoint):
            raise TypeError(
                "unsupported operand type(s) for +: '%s' and '%s'" %
                (type(self).__name__, type(other).__name__))
        return ScreenPoint(self._x - other._x, self._y - other._y)
    def __repr__(self):
        return "ScreenPoint(%s, %s)" % (self._x, self._y)

    
class PolarPoint(object):
    """
    A point in polar coordinates.

    The ``r`` and ``theta`` components of the point may be accessed as
    with any sequence::

       # pt is a ScreenPoint
       r = pt[0]
       theta = pt[1]

       # or

       r, theta = pt    
    """
    def __init__(self, r, theta):
        self._r = r
        self._theta = theta
    def __getitem__(self, key):
        if key == 0: return self._r
        elif key == 1: return self._theta
        else: raise IndexError(key)
    def __eq__(self, other):
        # XXX: Note that we're not doing float comparisons, these will
        # need to be exactly equal....
        if (isinstance(other, PolarPoint) and
            self._r == other._r and
            self._theta == other._theta):
            return True
    def __repr__(self):
        return "PolarPoint(%s, %s)" % (self._r, self._theta)


#
# Point conversion functions
#
    
def cartesian_from_polar(pt):
    """
    Converts a :class:`PolarPoint` *pt* into a :class:`CartesianPoint`
    measured relative to the polar coordinate's origin.
    """
    if not isinstance(pt, PolarPoint):
        raise TypeError(type(pt))
    r, theta = pt
    return CartesianPoint(r*math.cos(theta), r*math.sin(theta))

def polar_from_cartesian(pt):
    """
    Converts a :class:`CartesianPoint` *pt* into a :class:`PolarPoint`
    measured relative to the Cartesian point's origin.
    """
    if not isinstance(pt, CartesianPoint):
        raise TypeError(
            "input to polar_from_cartesian must be CartesianPoint")
    x, y = pt
    r = math.sqrt(x**2 + y**2)
    theta = math.atan2(y, x)
    if theta < 0:
        theta += 2*math.pi
    return PolarPoint(r, theta)

def screen_from_cartesian(pt, height):
    """
    Converts a :class:`CartesianPoint` *pt* to the corresponding
    :class:`ScreenPoint`. This is only possible if the y-axis is
    bounded; *height* gives the total size of the vertical dimension
    from the origin to the end of the plot space.
    """
    if not isinstance(pt, CartesianPoint):
        raise TypeError(
            "input to screen_from_cartesian must be CartesianPoint")
    x, y = pt
    return ScreenPoint(x, height - y)

def cartesian_from_screen(pt, height):
    """
    Convert a :class:`CartesianPoint` *pt* to the corresponding
    :class:`ScreenPoint`. This is only possible if the y-axis is
    bounded; *height* gives the total size of the vertical dimension
    from the origin to the end of the plot space.
    """
    if not isinstance(pt, ScreenPoint):
        raise TypeError("input to cartesian_from_screen must be ScreenPoint")
    x, y = pt
    return CartesianPoint(x, height - y)


#
# Rectangles and Spacing
#

class Spacing(object):
    """
    Interior padding within a :class:`Rectangle`, in device units. To
    access the top, bottom, left and right padding respectively,
    access this object as a tuple::
    
      s = Spacing(1,2,3,4)
      tpad = s[0] # 1
      bpad = s[1] # 2
      lpad = s[2] # 3
      rpad = s[3] # 4

      # or

      tpad, bpad, lpad, rpad = s
    """
    def __init__(self, tpad=0, bpad=0, lpad=0, rpad=0):
        self._spacing=(tpad, bpad, lpad, rpad)
    def __getitem__(self, idx):
        return self._spacing[idx]
    def __repr__(self):
        return "Spacing(tpad=%s, bpad=%s, lpad=%s, rpad=%s)" % (
            self.tpad, self.bpad, self.lpad, self.rpad)

class Rectangle(object):
    """
    A rectangle of the given *width* and *height*. The rectangle uses
    screen coordinates, so the if the top left corner is at ``(x,
    y)``, the bottom right corner is at ``(x + width, y + height)``.
    """
    def __init__(self, width, height):
        self._width = width
        self._height = height
    def __eq__(self, other):
        if (isinstance(other, Rectangle) and
            self._width  == other._width and
            self._height == other._height):
            return True
        else:
            return False
    def __repr__(self):
        return "Rectangle(%s, %s)" % (self._width, self._height)
    def get_width(self): return self._width
    def get_height(self): return self._height
    def pad(self, spacing):
        """
        Returns a :class:`Rectangle` object whose size has been
        reduced according to the values of *spacing*, and a
        :class:`ScreenPoint` object representing the offset between
        the old rectangle's top left corner and that of the new
        rectangle.

        *spacing* is a tuple ``tpad, bpad, lpad, rpad`` indicating the
        amount of top, bottom, left and right padding,
        respectively. Values are in device units.
        """
        if not isinstance(spacing, Spacing):
            raise TypeError("input to pad must be Spacing")
        tpad, bpad, lpad, rpad = spacing
        new_width = max(self._width - lpad - rpad, 0)
        new_height = max(self._height - tpad - bpad, 0)
        return Rectangle(new_width, new_height), ScreenPoint(lpad, tpad)
        
    def get_corners(self,
                    top_left=None,
                    top_right=None,
                    bottom_left=None,
                    bottom_right=None,
                    center=None):
        """
        Returns the corners of this :class:`Rectangle` object assuming
        one of its corners or center is anchored to a particular point.
 
        If one of *top_left*, *top_right*, *bottom_left*,
        *bottom_right* or *center* is specified, it is a
        :class:`ScreenPoint` object indicating the corner relative to
        which the rectangle is anchored. (If more than one is
        specified, the first one in the order above is used.) By
        default, the rectangle is assumed to be drawn with the top
        left corner at ``(0,0)``.
        """
        def typecheck(pt):
            if not isinstance(pt, ScreenPoint):
                raise TypeError("Input to Rectangle.get_corners must "
                                "be ScreenPoint")
        w, h = self._width, self._height

        rel_top_left     = ScreenPoint(0, 0)
        rel_top_right    = ScreenPoint(w, 0)
        rel_bottom_left  = ScreenPoint(0, h)
        rel_bottom_right = ScreenPoint(w, h)
        
        if top_left is not None:
            typecheck(top_left)
            top_left = top_left
        elif top_right is not None:
            typecheck(top_right)
            top_left = top_right - rel_top_right
        elif bottom_left is not None:
            typecheck(bottom_left)
            top_left = bottom_left - rel_bottom_left
        elif bottom_right is not None:
            typecheck(bottom_right)
            top_left = bottom_right - rel_bottom_right
        elif center is not None:
            # Always a weird one
            typecheck(center)
            top_left = center - ScreenPoint(w/2, h/2)
        else:
            top_left = ScreenPoint(0, 0)

        return tuple(map(lambda pt: top_left + pt,
                         (rel_top_left, rel_top_right,
                          rel_bottom_left, rel_bottom_right)))



    
__all__ = """
CartesianPoint
Point
PolarPoint
Rectangle
ScreenPoint
""".split()
