#  Copyright (C) 2012 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Rayon and related source code is subject to the terms of
#  the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""Markers used especially for marking ticks in a border."""
from rayon import RayonException
import markers
import text
import miscutils

default_labeler = text.Labeler()

default_marker_args = {
    "hline": {"before": 0, "after": 2},
    "vline": {"below":  0, "above": 2},
}

# NOTE: TickMarkers are currently undocumented and will probably be
# removed in Rayon 2.0. (The possible exception is ColumnTickMarker,
# which is used in iSiLK.)

#
# Base TickMarker classes
#

class TickMarker(markers.LabeledMarker):
    """
    A marker used to draw ticks along a chart border.

    *marker_type* is the style of :class:`~rayon.markers.Marker`
    object used to make the tick mark (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_marker`). *marker_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.markers.Marker` object as keyword arguments. (I.e.,
    equivalently to calling ``toolbox.new_marker(marker_type,
    **marker_args)``)

    If specified, *labeler* is a :class:`~rayon.text.Labeler` object
    used to draw the text labels.

    *label_position* describes the position of the label relative to
    the tick mark. Valid values are ``n`` (above), ``s`` (below),
    ``e`` (to the left) and ``w`` (to the right).

    *label_spacing* is the amount of space between the label and the
    marker, in device units (points or pixels).
    """
    pass

class HorizontalTickMarker(TickMarker):
    """
    A marker used to mark ticks along a horizontal (\"X\") axis.
    """
    def __init__(self, marker_type="vline",
                 marker_args=None,
                 labeler=default_labeler,
                 label_position="s", label_spacing=5):
        if marker_args is None:
            marker_args = default_marker_args.get(marker_type, {})
        marker = markers.create(marker_type, **marker_args)
        markers.LabeledMarker.__init__(self, marker, labeler,
                                       label_position, label_spacing)



class VerticalTickMarker(TickMarker):
    """
    A marker used to mark ticks along a vertical (\"Y\") axis.
    """
    def __init__(self, marker_type="hline",
                 marker_args=None,
                 labeler=default_labeler,
                 label_position="e", label_spacing=5):
        if marker_args is None:
            marker_args = default_marker_args.get(marker_type, {})
        marker = markers.create(marker_type, **marker_args)
        markers.LabeledMarker.__init__(self, marker, labeler,
                                       label_position, label_spacing)


#
# "Simple" (i.e., simplified) tick markers
#


class SimpleHorizontalTickMarker(HorizontalTickMarker):
    """
    A :class:`HorizontalTickMarker` whose interface is a set of simple
    Python values. In practice, this class is often simpler to create
    and use than a normal :class:`HorizontalTickMarker` because no
    additional objects must be created.

    *marker_type* is the style of :class:`~rayon.markers.Marker`
    object used to make the tick mark (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_marker`). *marker_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.markers.Marker` object as keywork arguments. (I.e.,
    equivalently to calling ``toolbox.new_marker(marker_type,
    **marker_args)``)

    *font_size*, *font_family* and *font_style* define the font used
    for the tick label, and take the values described in
    :ref:`specifying-text-properties`. *halign* and *valign* specify
    the horizontal and vertical alignment of the label.

    *color* is the color of the label text. *bgcolor* is the color of
    the label background. *angle* is the angle of the text (0 ==
    horizontal, left to right).

    If specified, *border* is a text line specification (see
    :ref:`specifying-lines`) describing the :class:`~rayon.lines.Line`
    object to use to draw border lines around the label text's
    bounding box. *border_left*, *border_right*, *border_top* and
    *border_bottom* specify lines that will only be drawn on the left,
    right, top and bottom edges of the box, respectively. If *border*
    is specified, the other *border_* arguments will be ignored.

    *label_position* describes the position of the label relative to
    the tick mark. Valid values are ``n`` (above), ``s`` (below),
    ``e`` (to the left) and ``w`` (to the right).

    *label_spacing* is the amount of space between the label and the
    marker, in device units (points or pixels).
    """
    def __init__(self, marker_type="vline",
                 marker_args=None,
                 font_size="normal", font_family="default",
                 font_style="normal", font_weight="normal",
                 halign="center", valign="top",
                 color=(0, 0, 0), bgcolor=(0, 0, 0, 0), angle=0,
                 border=None, border_top=None, border_bottom=None,
                 border_left=None, border_right=None,
                 label_position="s", label_spacing=5):
        labeler = text.Labeler(font_size=font_size,
                               font_family=font_family,
                               font_style=font_style,
                               font_weight=font_weight,
                               halign=halign, valign=valign,
                               color=color, bgcolor=bgcolor,
                               angle=angle,
                               border=border,
                               border_top=border_top,
                               border_bottom=border_bottom,
                               border_left=border_left,
                               border_right=border_right)
        HorizontalTickMarker.__init__(self, marker_type, marker_args,
                                      labeler, label_position, label_spacing)



class SimpleVerticalTickMarker(VerticalTickMarker):
    """
    A :class:`VerticalTickMarker` whose interface is a set of simple
    Python values. In practice, this class is often simpler to create
    and use than a normal :class:`VerticalTickMarker` because no
    additional objects must be created.

    *marker_type* is the style of :class:`~rayon.markers.Marker`
    object used to make the tick mark (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_marker`). *marker_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.markers.Marker` object as keywork arguments. (I.e.,
    equivalently to calling ``toolbox.new_marker(marker_type,
    **marker_args)``)

    *font_size*, *font_family* and *font_style* define the font used
    for the tick label, and take the values described in
    :ref:`specifying-text-properties`. *halign* and *valign* specify
    the horizontal and vertical alignment of the label.

    *color* is the color of the label text. *bgcolor* is the color of
    the label background. *angle* is the angle of the text (0 ==
    horizontal, left to right).

    If specified, *border* is a text line specification (see
    :ref:`specifying-lines`) describing the :class:`~rayon.lines.Line`
    object to use to draw border lines around the label text's
    bounding box. *border_left*, *border_right*, *border_top* and
    *border_bottom* specify lines that will only be drawn on the left,
    right, top and bottom edges of the box, respectively. If *border*
    is specified, the other *border_* arguments will be ignored.

    *label_position* describes the position of the label relative to
    the tick mark. Valid values are ``n`` (above), ``s`` (below),
    ``e`` (to the left) and ``w`` (to the right).

    *label_spacing* is the amount of space between the label and the
    marker, in device units (points or pixels).
    """
    def __init__(self, marker_type="hline",
                 marker_args=None,
                 font_size="normal", font_family="default",
                 font_style="normal", font_weight="normal",
                 halign="right", valign="center",
                 color=(0, 0, 0), angle=0,
                 bgcolor=(0, 0, 0, 0),
                 border=None, border_top=None, border_bottom=None,
                 border_left=None, border_right=None,
                 label_position="w", label_spacing=5):
        labeler = text.Labeler(font_size, font_family, font_style, font_weight,
                                halign, valign, color, bgcolor, angle, border,
                                border_top, border_bottom,
                                border_left, border_right)
        VerticalTickMarker.__init__(self, marker_type, marker_args,
                                    labeler, label_position, label_spacing)



#
# Column/Row tick markers
#

class ColumnTickMarker(VerticalTickMarker):
    """A L{VerticalTickMarker} with a L{ColumnLabeler} to align
    multipart labels into columns."""
    def __init__(self, labelers, marker_type="hline",
                 marker_args=None,
                 box_width=1, col_width="equal",
                 tick_data=None,
                 halign="left", valign="center", gutter=0,
                 border=None, border_top=None, border_bottom=None,
                 border_left=None, border_right=None,
                 label_position="e", label_spacing=5):
        """Constructor.

        @type  labelers: sequence
        @param labelers: L{text.Labeler} objects to be used to labeel
            individual columns

        @param marker_type: type of marker to create. See
            L{markers.create} for valid type names.

        @param box_width: see L{text.ColumnLabeler}

        @param col_width: see L{text.ColumnLabeler}. An additional
            input is available, the string literal C{fromdata}. If
            specified, C{ColumnTickMarker} will compute column widths
            sufficient to contain all the data in I{tick_data}, which
            must be specified.

        @type  tick_data: sequence of 2-tuples
        @param tick_data: tick data used as input to the
            L{ticks.Ticks} object.

        @param halign: see L{text.ColumnLabeler}
        @param valign: see L{text.ColumnLabeler}
        @param gutter: see L{text.ColumnLabeler}
        @param border: see L{text.ColumnLabeler}
        @param border_top: see L{text.ColumnLabeler}
        @param border_bottom: see L{text.ColumnLabeler}
        @param border_left: see L{text.ColumnLabeler}
        @param border_right: see L{text.ColumnLabeler}

        @param label_position: see L{markers.LabeledMarker}
        @param label_spacing: see L{markers.LabeledMarker}
        """
        # Make a set of functions that calculate column width properly
        # just prior to drawing
        if col_width == "fromdata":
            if tick_data is None:
                raise RuntimeError("tick_data is required if "
                                   "col_width=\"fromdata\"")
            # Convert to column-major
            columns = zip(*(text for ticks, text
                            in miscutils.tick_data_iter(tick_data)))
            def mk_width_function(l, col):
                def width_function(ctx, width):
                    # Yup, ignoring width. 
                    width = 0
                    for i in col:
                        bbox = l.get_bounding_box(ctx, i)
                        box_width = abs(bbox[0][0] - bbox[2][0])
                        width = max(width, box_width)
                    return width
                return width_function
            col_width = [mk_width_function(labeler, column)
                         for labeler, column in zip(labelers, columns)]
        labeler = text.ColumnLabeler(labelers, box_width, col_width,
                                     halign, valign, gutter,
                                     border, border_top, border_bottom,
                                     border_left, border_right)
        VerticalTickMarker.__init__(self, marker_type, marker_args,
                                    labeler, label_position, label_spacing)
                                    

# TODO: A RowTickMarker to correspond with the RowLabeler, when/if
# that object is created.

simple_tickmarker_types = {
    "vertical": SimpleVerticalTickMarker,
    "horizontal": SimpleHorizontalTickMarker
}

tickmarker_types = {
    "vertical": VerticalTickMarker,
    "horizontal": HorizontalTickMarker
}

# Not including ColumnTickMarker. (1) It's going away, and (2) if you
# do want it, you should know what you're doing enough to import it
# manually.
def create(tm_type, **kargs):
    if kargs.has_key('labeler'):
        return tickmarker_types[tm_type](**kargs)
    else:
        return simple_tickmarker_types[tm_type](**kargs)

def list_tickmarker_types():
    return tuple((k, v) for k, v in tickmarker_types.items())


__all__ = ["HorizontalTickMarker", "VerticalTickMarker",
           "SimpleHorizontalTickMarker", "SimpleVerticalTickMarker",
           "ColumnTickMarker", "create"]
